#############################################################################
##
#F  SagGroupOps.FCoveringGroup( G, form ) . . . .<form>-covering group of <G>
##
SagGroupOps.FCoveringGroup := function( G, form )

    local U,            # approximate cover
          system,       # ag system for normalizer
          head, tail,   # of nilpotent factor
          hhead,        # next nilpotent factor
          locals,       # local residuals
          ffirst,       # correction factor
          first, next,  # of layer
          p,            # prime
          primes,       # list of primes
          reflayer,     # refined layer
          index,        # to cut out eccentric elements
          j, k,         # positions 
          i;            # loop variable
    
    # set up the cover
    U      := Copy( G );

    # set up system record
    head := G.head[2];
    if IsBound( form.support ) then
        index := Filtered( [1..head-1], 
                            x -> G.weights[x][3] in form.support );
    else
        index := [1..head-1];
    fi;
    system          := rec();
    system.H        := G;
    system.base     := G.generators{ index };
    system.weights  := G.weights{ index };
    system.Fcentral := List( index, x -> true );
    system.Ffirst   := Filtered( G.first{[ 1..Position( G.first, head) ]},
                                 x -> x in index );

    InfoForm2("#I 1 head done \n");

    # run down lower nilpotent series (the first factor is trivial)
    for i  in [ 2..Length(G.head)-1 ]  do
        head  := G.head[ i ];
        tail  := G.tail[ i ];
        hhead := G.head[ i+1 ];

        # get local residuals
        primes := Set( List( G.weights{[head..tail-1]}, x -> x[3] ) );
        locals := List( primes, x -> form.LocalResidual( U, x ) );

        # compute correction factor
        ffirst := head - Length( system.base ) - 1;

        # refine nilpotent layer
        j := Position( G.first, head );
        while G.first[j] < hhead do
            first := G.first[j];
            next  := G.first[j+1];
            p := G.weights[ first ][3];

            if not IsBound( form.support ) or
               (IsBound( form.support ) and p in form.support) then
                k := Position( primes, p );
                reflayer := G.operations.RefinedBaseLayer( G, locals[k], j );
                Append( system.base, reflayer.base );
                Append( system.Fcentral, reflayer.central );
                Append( system.Ffirst, 
                        List( reflayer.firsts, x -> x - ffirst ) );
                Append( system.weights, G.weights{[ first..next-1 ]} );
            else
                Append( system.base, G.generators{[ first..next-1 ]} );
                Append( system.Fcentral, List( [first..next-1], x -> false ));
                Append( system.Ffirst, [next - ffirst] );
            fi;

            j := j + 1;
        od;
        system.depths := List( system.base, DepthAgWord );

        # compute layers relative to first for full base
        system.Flayers := List( system.base, x -> false );
        j := 1;
        for k in [1..Length( system.base )] do
            if system.Ffirst[j] = k then
                j := j + 1;
            fi;
            system.Flayers[k] := j-1;
        od;

        # calculate p-complements of residuals
        for j in [1..Length(primes)] do
            if not IsBound( form.support ) or 
               (IsBound( form.support ) and primes[j] in form.support) then
                locals[j] := G.operations.PComplement(G, locals[j], primes[j]);
            fi;
        od;

        j := Length( system.base );
        while j >= 1  and system.Fcentral[ j ]  do
            j := j - 1;
        od;
        while j >= 1  do
            if system.Fcentral[ j ]  then
                p := system.weights[ j ][ 3 ];
                k := Position( primes, p );
                if not IsBool( k ) then
                    G.operations.ChangeGenerator( system, locals[ k ], j );
                fi;
            fi;
            j := j - 1;
        od;

        # set up for next head
        index := Filtered( [1..Length(system.Fcentral)], 
                           x -> system.Fcentral[x] );
        system.base := system.base{ index };
        system.weights := system.weights{ index };
        system.Fcentral := List( index, x -> true );

        system.Ffirst  := [ ];
        k := 1;
        for j in [1..Length(system.base)] do
            if DepthAgWord( system.base[j] ) >= G.first[k] then
                k := k + 1;
                Add( system.Ffirst, j );
            fi;
        od;
        Add( system.Ffirst, Length( system.base ) + 1 );
            
        U := G.operations.Subgroup( G, Concatenation( system.base, 
             G.generators{[ hhead..Length(G.generators) ]} ) );

        InfoForm2("#I ",i," head done \n");
    od;
    return U;
end;

#############################################################################
##
#F  AgGroupOps.FCoveringGroup( G, form ) . . . . <form>-covering group of <G>
##
AgGroupOps.FCoveringGroup := function( G, form )
    local S, N;
    if Length( G.generators ) = 0 then return G; fi;
    InfoForm2("#I compute special ag group \n");
    S := SpecialAgGroup( G );
    InfoForm2("#I compute normalizer \n");
    N := S.operations.FCoveringGroup( S, form );
    return AgSubgroup( G, List(Igs(N), x -> Image(S.bijection, x)), false );
end;

#############################################################################
##
#F  FCoveringGroup( G, form ) . . . . . . . . . .<form>-covering group of <G>
##
FCoveringGroup := function( G, form )
    local func, str, i;

    # check argruments
    if not IsAgGroup( G ) or not IsFormation( form )  then
        Error(" Usage: FCoveringGroup( <G>, <formation> )\n");
    fi;
    if not IsBound( form.LocalResidual )  then
        Error(" Usage: formation needs to be locally defined \n");
    fi;
    if not IsBound( form.isIntegrated ) or not form.isIntegrated  then
        Error(" Usage: <formation> needs to be integrated \n");
    fi;

    # set up record structure 
    if not IsBound( G.formation ) then
        G.formation := rec( );
    fi;

    if not IsBound( form.support ) then

        if not IsBound( G.formation.(form.name) ) then 
            G.formation.(form.name) := rec( );
        fi;

        # check whether we know result
        if not IsBound( G.formation.(form.name).FCoveringGroup ) then

            # use the right function
            if IsBound( form.FCoveringGroup ) then
                G.formation.(form.name).FCoveringGroup := 
                                        form.FCoveringGroup( G );
            else
                G.formation.(form.name).FCoveringGroup := 
                             G.operations.FCoveringGroup( G, form );
            fi;
        fi;

        # return
        return G.formation.(form.name).FCoveringGroup;
    else
        
        # turn support into string
        str := ConcatenationString( "Support", String( form.support[1] ) );
        for i in [2..Length( form.support )] do
            str := ConcatenationString( str, "_", String( form.support[i] ) );
        od;
        str := ConcatenationString( form.name, str );

        # add another record component
        if not IsBound( G.formation.(str) ) then
            G.formation.(str) := rec();
        fi;

        # check whether result is known
        if not IsBound( G.formation.(str).FCoveringGroup ) then

            # use the right function
            if IsBound( form.FCoveringGroup ) then
                G.formation.(str).FCoveringGroup := 
                                        form.FCoveringGroup( G );
            else
                G.formation.(str).FCoveringGroup := 
                            G.operations.FCoveringGroup( G, form );
            fi;
        fi;

        # return
        return G.formation.(str).FCoveringGroup;
    fi;
end;

#############################################################################
##                         Second strategy                                 ##
############################################################################# 

#############################################################################
##
#F  SagGroupOps.FCoveringGroupR( G, formation ) . . . . . . . . 2nd strategie
##
SagGroupOps.FCoveringGroupR := function( G, form )

    local U,            # the covering group
          max,          # composition length of G
          head,         # the first nilpotent factor of G
          system,       # ag system of cover
          first, next,  # of a layer
          wt, p,        # corresponding weight and prime
          N, M,         # corresponding normal subgroups
          loc,          # local residual 
          comp, ncomp,  # p-complement and O^p of loc
          residual,     # residual modulo M
          gens,         # some new ag system elements
          layer,        # number of layer
          Ffirst,       # first of layer
          q,            # prime of element
          i, j, k;      # indices
    
    # set up the cover
    U      := Copy( G );
    max    := Length( G.generators );

    # set up system record
    head := G.head[2];
    system := rec();
    system.H := G;
    system.base := G.generators{[ 1..head-1 ]};
    system.Flayers  := G.layers{[ 1..head-1 ]};
    system.Ffirst   := G.first{[ 1..Position( G.first, head) ]};
    system.weights  := G.weights{[ 1..head-1 ]};
    if IsBound( form.support ) then
        system.Fcentral := List( [1..head-1], 
                                 x -> G.weights[x][3] in form.support );
    else
        system.Fcentral := List( [1..head-1], x -> true );
    fi;

    InfoForm2("#I 1 head done \n");

    # run down elementary abelian series (the first factor is trivial)
    for i  in [ Position( G.first, head )..Length(G.first)-1 ]  do
        first := G.first[ i ];
        next  := G.first[ i+1 ];
        wt    := G.weights[ first ];
        p     := wt[ 3 ];
        if not IsBound( form.support) or
           (IsBound( form.support ) and p in form.support) then

            N := G.operations.AgSubgroup( G, G.generators{[first..max]}, true);
            M := G.operations.AgSubgroup( G, G.generators{[next..max]}, true);

            # get local residual
            loc   := form.LocalResidual( U, p );

            # compute p-complement and O^p
            comp  := G.operations.PComplement( G, loc, p );
            ncomp := loc.operations.NormalClosure( loc, comp );

            # compute residual
            residual := FCommutatorSubgroup( comp, N, M );

            # update system record for normalizer
            gens := FactorArg( N, residual ).generators;
            if Length( gens ) > 0 then
                Append( system.base, gens );
                Append( system.Fcentral, List( gens, x -> true ) );
                Append( system.weights, List( gens, x -> wt ) );
                layer := system.Flayers[ Length( system.Flayers ) ];
                Append( system.Flayers, List( gens, x -> layer + 1 ) );
                Ffirst := system.Ffirst[Length(system.Ffirst)] + Length(gens);
                Add( system.Ffirst, Ffirst );
            fi;

            gens := FactorArg( residual, M ).generators;
            if Length( gens ) > 0 then
                Append( system.base, gens );
                Append( system.Fcentral, List( gens, x -> false ) );
                Append( system.weights, List( gens, x -> wt ) );
                layer := system.Flayers[ Length( system.Flayers ) ];
                Append( system.Flayers, List( gens, x -> layer + 1 ) );
                Ffirst := system.Ffirst[Length(system.Ffirst)] + Length(gens);
                Add( system.Ffirst, Ffirst );

                # modify central elements
                j := system.Ffirst[ Length( system.Ffirst ) - 1 ] - 1;
                while j >= 1  do
                    q := system.weights[ j ][ 3 ];
                    if p = q then
                        G.operations.ChangeGenerator( system, comp, j );
                    fi;
                    j := j - 1;
                od;

                # cut out eccentric part
                Unbind( system.Ffirst[ Length( system.Ffirst ) ] );
                j := system.Ffirst[ Length( system.Ffirst ) ] - 1;
                system.base := system.base{[1..j]};
                system.Fcentral := system.Fcentral{[1..j]};
                system.weights := system.weights{[1..j]};
                system.Flayers := system.Flayers{[1..j]};

                # change U
                U := G.operations.AgSubgroup( G, 
                     Concatenation( system.base, G.generators{[ next..max ]} ),
                     false );
            fi;
        fi;
        InfoForm2("#I ",i," layer done \n");
    od;
    return U;
end;

#############################################################################
##
#F  AgGroupOps.FCoveringGroupR( G, form ) . . . . . . . . . . . 2nd strategie
##
AgGroupOps.FCoveringGroupR := function( G, form )
    local S, N;
    if Length( G.generators ) = 0 then return G; fi;
    InfoForm2("#I compute special ag group \n");
    S := SpecialAgGroup( G );
    InfoForm2("#I compute normalizer \n");
    N := S.operations.FCoveringGroupR( S, form );
    return AgSubgroup( G, List(Igs(N), x -> Image(S.bijection, x)), false );
end;

#############################################################################
##
#F  FCoveringGroupR( G, form ) . . . . . . . . . . . . . . . . .2nd strategie
##
FCoveringGroupR := function( G, form )
    local func, str, i;

    # check argruments
    if not IsAgGroup( G ) or not IsFormation( form )  then
        Error(" Usage: FCoveringGroupR( G, formation )\n");
    fi;
    if not IsBound( form.isIntegrated ) or not form.isIntegrated  then
        Error(" Usage: <formation> needs to be integrated \n");
    fi;

    # set up record structure 
    if not IsBound( G.formation ) then
        G.formation := rec( );
    fi;

    if not IsBound( form.support ) then

        if not IsBound( G.formation.(form.name) ) then 
            G.formation.(form.name) := rec( );
        fi;

        # check whether we know result
        if not IsBound( G.formation.(form.name).FCoveringGroup ) then

            # use the right function
            if IsBound( form.FCoveringGroup ) then
                G.formation.(form.name).FCoveringGroup := 
                                        form.FCoveringGroup( G );
            else
                G.formation.(form.name).FCoveringGroup := 
                                        G.operations.FCoveringGroupR( G, form );
            fi;
        fi;

        # return
        return G.formation.(form.name).FCoveringGroup;
    else
        
        # turn support into string
        str := ConcatenationString( "Support", String( form.support[1] ) );
        for i in [2..Length( form.support )] do
            str := ConcatenationString( str, "_", String( form.support[i] ) );
        od;
        str := ConcatenationString( form.name, str );

        # add another record component
        if not IsBound( G.formation.(str) ) then
            G.formation.(str) := rec();
        fi;

        # check whether result is known
        if not IsBound( G.formation.(str).FCoveringGroup ) then

            # use the right function
            if IsBound( form.FCoveringGroup ) then
                G.formation.(str).FCoveringGroup := 
                                        form.FCoveringGroup( G );
            else
                G.formation.(str).FCoveringGroup := 
                                        G.operations.FCoveringGroupR( G, form );
            fi;
        fi;

        # return
        return G.formation.(str).FCoveringGroup;
    fi;
end;

