#############################################################################
##
#F  NilpotentResidual( <G> ) . . . . . . . . . . . .nilpotent residual of <G>
##
NilpotentResidual := function( G )
    if IsBound( G.nilpotentResidual ) then
        return G.nilpotentResidual;
    fi;
    G.nilpotentResidual := G.operations.NilpotentResidual( G );
    return G.nilpotentResidual;
end;

SagGroupOps.NilpotentResidual := function( G )
    return G.operations.AgSubgroup( G, 
           G.generators{ [ G.head[2] .. Length( G.generators ) ] }, true );
end;

AgGroupOps.NilpotentResidual := function( G )

    local  base, gens, g, p, h, q;
 
    Cgs( G );

    # build up the necessary generators
    base := List( G.cgs, x -> PPart( G, x, RelativeOrderAgWord( x ) ) );
    gens := [ ];
    for g  in base  do
        p := RelativeOrderAgWord( g );
        for h  in base  do
            q := RelativeOrderAgWord( h );
            if q <> p  then
                Add( gens, Comm( g, h ) );
            fi;
        od;
    od;

    # close under operation
    return G.operations.NormalClosure( G, 
           G.operations.Subgroup( G, gens ) );
end;

#############################################################################
##
#F  PResidual( <G>, <prime> ) . . . . . . . . . . . . <prime>-residual of <G>
##
PResidual := function( G, prime )
    if IsBound( G.piResidual ) and IsBound( G.piResidual[prime] ) then
        return G.piResidual[prime];
    elif not IsBound( G.piResidual ) then
        G.piResidual := [ ];
    fi;
    G.piResidual[prime] := G.operations.PResidual( G, prime );
    return G.piResidual[prime];
end;

SagGroupOps.PResidual := function( G, prime )
    local gens, i;
    gens := [ ];
    for i in [ 1 .. Length( G.generators ) ]  do
        if G.weights[ i ][ 1 ] > 1 or G.weights[ i ][ 3 ] <> prime  then
            Add( gens, G.generators[ i ] );
        fi;
    od;
    return G.operations.AgSubgroup( G, gens, true );
end;

AgGroupOps.PResidual := function( G, prime )

    local base, gens, g, p, i;

    Cgs( G );
    gens := [ ];
    for g  in G.cgs  do
        p := RelativeOrderAgWord( g );
        if p <> prime then
            Add( gens, PPart( G, g, p ) );
        fi;
    od;
    return G.operations.NormalClosure( G, G.operations.Subgroup( G, gens ) );
end;

#############################################################################
##
#F  PiResidual( <G>, <primes> ) . . . . . . . . . . . <primes>-resiual of <G>
##
PiResidual := function( G, primes )
    local prod;
    prod := Product( primes );
    if IsBound( G.piResidual ) and IsBound( G.piResidual[prod] ) then
        return G.piResidual[prod];
    elif not IsBound( G.piResidual ) then
        G.piResidual := [ ];
    fi;
    G.piResidual[prod] := G.operations.PiResidual( G, primes );
    return G.piResidual[prod];
end;

AgGroupOps.PiResidual := function( G, primes )

    local base, gens, g, p, i;

    Cgs( G );
    gens := [ ];
    for g  in G.cgs  do
        p := RelativeOrderAgWord( g );
        if not p in primes then
            Add( gens, PPart( G, g, p ) );
        fi;
    od;
    return G.operations.NormalClosure( G, G.operations.Subgroup( G, gens ) );
end;

SagGroupOps.PiResidual := AgGroupOps.PiResidual;

#############################################################################
##
#F  CoprimeResidual( <G>, <primes> ) . . . . . . . .<primes>'-residual of <G>
##
CoprimeResidual := function( G, primes )
    local prod;
    prod := Product( primes );
    if IsBound( G.coprimeResidual ) and IsBound( G.coprimeResidual[prod] ) 
    then
        return G.coprimeResidual[prod];
    elif not IsBound( G.coprimeResidual ) then
        G.coprimeResidual := [ ];
    fi;
    G.coprimeResidual[prod] := G.operations.CoprimeResidual( G, primes );
    return G.coprimeResidual[prod];
end;

AgGroupOps.CoprimeResidual := function( G, primes )

    local base, gens, g, p, i;

    Cgs( G );
    gens := [ ];
    for g  in G.cgs  do
        p := RelativeOrderAgWord( g );
        if p in primes then
            Add( gens, PPart( G, g, p ) );
        fi;
    od;
    return G.operations.NormalClosure( G, G.operations.Subgroup( G, gens ) );
end;

SagGroupOps.CoprimeResidual := AgGroupOps.CoprimeResidual;

#############################################################################
##
#F  ElementaryAbelianResidual( <G> ) . . . .abelian with square free exponent
##
ElementaryAbelianResidual := function( G )

    if IsBound( G.elementaryAbelianResidual ) then
        return G.elementaryAbelianResidual;
    fi;
    G.elementaryAbelianResidual := G.operations.ElementaryAbelianResidual(G);
    return G.elementaryAbelianResidual;
end;
    
SagGroupOps.ElementaryAbelianResidual := function( G )
    return G.operations.AgSubgroup( G, 
           G.generators{[G.tail[1]..Length(G.generators)]}, true );
end;

AgGroupOps.ElementaryAbelianResidual := function( G )
    local D, gens;
    D    := DerivedSubgroup(G);
    gens := List( FactorArg( G, D ).generators, 
                  x -> x ^ RelativeOrderAgWord( x ) );
    return G.operations.MergedIgs( G, D, gens, false );
end;

#############################################################################
##
#F  AbelianExponentResidual( <G>, <exp> ) . . . . . . .abelian with <exp>-law
##
##  Calculate the smallest normal subgroup <N> of <G> such that <G> / <N> is 
##  abelian and its exponents divides <exp>.
## 
AbelianExponentResidual := function( G, p )
    if not IsBound( G.abelianExponentResidual )  then
        G.abelianExponentResidual := [ ];
    fi;
    if IsBound( G.abelianExponentResidual ) and 
       IsBound( G.abelianExponentResidual[ p ] ) then
        return G.abelianExponentResidual[ p ];
    fi;
    G.abelianExponentResidual[ p ] := 
      G.operations.AbelianExponentResidual( G, p );
    return G.abelianExponentResidual[ p ];
end;

AgGroupOps.AbelianExponentResidual := function( G, p )
    local  D, gens;
    D := DerivedSubgroup( G );
    gens := List( FactorArg( G, D ).generators, x -> x ^ (p-1) );
    return G.operations.MergedIgs( G, D, gens, false );
end;
 
SagGroupOps.AbelianExponentResidual := AgGroupOps.AbelianExponentResidual;


#############################################################################
##
#F  Formation( name (, prime(s) ) ) . . . . . . . . . .examples of formations
##
Formation := function( arg )

    local formation, name, p, i;

    # check arguments 
    if not IsString( arg[1] ) then
        Error(" Usage: Formation( <name> (,<prime(s)>) ) \n");
    fi;
    if not (arg[1] = "Nilpotent"
       or   arg[1] = "Supersolvable" 
       or   arg[1] = "Abelian"
       or   arg[1] = "ElementaryAbelian"
       or   arg[1] = "PiGroups"
       or   arg[1] = "PNilpotent"
       or   arg[1] = "PLengthOne" )  then
        Error("Usage: name is not in list of formations \n");
    fi;
    if arg[1] = "PiGroups" then
       if not IsBound( arg[2] ) or not IsList( arg[2] ) then
           Error("Usage: You need to specify primes \n");
       fi;
    fi;
    if arg[1] = "PLengthOne" or arg[1] = "PNilpotent" then
       if not IsBound( arg[2] ) or not IsInt( arg[2] ) then
           Error("Usage: You need to specify prime \n");
       fi;
    fi;

    # set up formation record
    formation := rec( isFormation  := true, 
                      operations   := FormationOps );

    # set up name
    if arg[1] = "PNilpotent" then
        name := ConcatenationString( String(arg[2]), "Nilpotent" );
    elif arg[1] = "PLengthOne" then
        name := ConcatenationString( String(arg[2]), "LengthOne" );
    elif arg[1] = "PiGroups" then
        name := String( arg[2][1] );
        for i in [2..Length( arg[2] )] do
            name := ConcatenationString( name, "_", String( arg[2][i] ) );
        od;
        name := ConcatenationString( name, "Groups" );
    else 
        name := arg[1];
    fi;
    formation.name := name;

    # compute LocalResidual and / or FResidual
    if arg[1] = "Nilpotent" then
        formation.LocalResidual := function( G, p) return G; end;
        formation.FResidual     := NilpotentResidual;
        formation.isIntegrated  := true;
    elif arg[1] = "Supersolvable" then
        formation.LocalResidual := AbelianExponentResidual;
        formation.isIntegrated  := true;
    elif arg[1] = "Abelian" then
        formation.FResidual := DerivedSubgroup;
    elif arg[1] = "ElementaryAbelian" then
        formation.FResidual := ElementaryAbelianResidual;
    elif arg[1] = "PNilpotent" then
        formation.LocalResidual := function( G, p )
            local N;
            N := G.operations.PResidual( G, arg[2] );
            if p = arg[2] then
                return N;
            else
                return N.operations.CoprimeResidual( N, [arg[2]] );
            fi;
        end;
        formation.isIntegrated  := true;
        formation.FResidual := function( G )
            local N;
            N := G.operations.PResidual( G, arg[2] );
            return N.operations.CoprimeResidual( N, [arg[2]] );
        end;
    elif arg[1] = "PiGroups" then
        formation.LocalResidual := function( G, p )
            if p in arg[2] then
                return G.operations.PiResidual( G, arg[2] );
            else
                return [];
            fi;
        end;
        formation.isIntegrated  := true;
        formation.FResidual := function( G )
            return G.operations.PiResidual( G, arg[2] );
        end;
        formation.support := arg[2];
    elif arg[1] = "PLengthOne" then
        formation.LocalResidual := function( G, p )
            local N;
            N := G.operations.CoprimeResidual( G, [arg[2]] );
            if p = arg[2] then
                return N;
            else
                N := N.operations.PResidual( N, arg[2] );
                return N.operations.CoprimeResidual( N, [arg[2]] );
            fi;
        end;
        formation.isIntegrated  := true;
        formation.FResidual := function( G )
            local N;
            N := G.operations.CoprimeResidual( G, [arg[2]] );
            N := N.operations.PResidual( N, arg[2] );
            return N.operations.CoprimeResidual( N, [arg[2]] );
        end;
    fi;
    return formation;
end;



