#############################################################################
##
#F  Operations record for formations
##
FormationOps := rec();

#############################################################################
##
#F  FormationOps.\= ( <form1>, <form2> ) . . . . . . . . . . . .test equality
##
FormationOps.\= := function( form1, form2 )
    if IsBound( form1.support ) and IsBound( form2.support ) then
        return (form1.name = form2.name and form1.support = form2.support);
    elif not IsBound( form1.support ) and not IsBound( form2.support ) then
        return (form1.name = form2.name );
    else
        return false;
    fi;
end;
    
#############################################################################
##
#F  FormationOps.Print( <form> ) . . . . . . . . . . . . . . .print formation
##
FormationOps.Print := function( form )
    Print("formation of ",form.name," groups ");
    if IsBound( form.support ) then
        Print("with support ", form.support );
    fi;
end;

#############################################################################
##
#F  FIntersection( form1, form2 ) . . . . . . . . .intersection of formations
##
FIntersection := function( form1, form2 )

    # check arguments
    if not IsFormation( form1 ) or not IsFormation( form2 )  then
        Error("Sorry, need two formations to intersect\n");
    fi;

    # return
    return form1.operations.FIntersection( form1, form2 );
end;

FormationOps.FIntersection  := function( form1, form2 )
    local form;
 
    # set up
    form := rec(
    operations := FormationOps,
    isFormation := true );

    # get the name
    if form1.name <= form2.name  then
        form.name := ConcatenationString( form1.name, "And", form2.name );
    else
        form.name := ConcatenationString( form2.name, "And", form1.name );
    fi;

    # check integrated
    if IsBound( form1.isIntegrated ) and IsBound( form2.isIntegrated ) 
    then
        form.isIntegrated := form1.isIntegrated and form2.isIntegrated;
    fi;

    # if both have a LocalResidual then the intersection has as well
    if IsBound( form1.LocalResidual ) and IsBound( form2.LocalResidual )  
    then
        form.LocalResidual := function( G, p )
            local L1, L2;
            L1 := form1.LocalResidual( G, p );
            L2 := form2.LocalResidual( G, p );
            if not IsList( L1 ) and not IsList( L2 ) then
                return G.operations.MergedIgs( G, L1, Igs( L2 ), false );
            else
                return [];
            fi;
        end;
    fi;

    # if one has a FResidual then the intersection should have as well
    if IsBound( form1.FResidual ) or IsBound( form2.FResidual ) then
        form.FResidual := function( G )
            local R1, R2;
            R1 := FResidual( G, form1 );
            R2 := FResidual( G, form2 );
            return G.operations.MergedIgs( G, R1, Igs( R2 ), false );
        end;
   fi;

   # now add the support
   if IsBound( form1.support ) and IsBound( form2.support ) then
       form.support := Intersection( form1.support, form2.support );
   elif IsBound( form1.support ) then
       form.support := form1.support;
   elif IsBound( form2.support ) then
       form.support := form2.support;
   fi;

   return form;
end;

#############################################################################
##
#F  FProduct( form1, form2 ) . . . . . . . . . . . . . .product of formations
##
FProduct := function( form1, form2 )

    # check arguments
    if not IsFormation( form1 ) or not IsFormation( form2 )  then
        Error("Sorry, need two formations to build product \n");
    fi;

    # return
    return form1.operations.FProduct( form1, form2 );
end;

FormationOps.FProduct := function( form1, form2 )
    local form;
     
    # set up
    form := rec(
    name := ConcatenationString( form1.name, "By", form2.name ),
    operations := FormationOps,
    isFormation := true );

    # check integrated
    if IsBound( form2.isIntegrated ) then
        form.isIntegrated := form2.isIntegrated;
    fi;

    # compute LocalResidual if possible
    if IsBound( form2.LocalResidual ) then
       form.LocalResidual := function( G, p )
           local R, L;
           R := FResidual( G, form1 );
           L := form2.LocalResidual( R, p );
           if not IsList( L ) then
               return L;
           else
               return [];
           fi;
       end;
    fi;

    # compute FResidual is necessary
    if IsBound( form2.FResidual ) then
        form.FResidual := function( G )
            local R;
            R := FResidual( G, form1 );
            return form2.FResidual( R );
        end;
    fi;

    # add support
    if IsBound( form1.support ) and IsBound( form2.support ) then
        form.support := Union( form1.support, form2.support );
    fi;

    return form;
end;

#############################################################################
##
#F  FIntegrated( form ) . . . . . . . . . . . . .compute integrated formation
##
FIntegrated := function( form )

    # check argument
    if not IsFormation( form )  then
        Error("Sorry, need a formation to integrate \n");
    elif not IsBound( form.LocalResidual ) then
        Error("Sorry, formation has to be locally defined \n");
    fi;

    # check whether formation is integrated
    if IsBound( form.isIntegrated ) and form.isIntegrated then
        return form;
    fi;
 
    # if result is unknown, compute it
    if not IsBound( form.integratedFormation ) then
        form.integratedFormation := form.operations.FIntegrated( form );
    fi;

    return form.integratedFormation;
end;

FormationOps.FIntegrated := function( form )

    local newform;

    # set up new formation
    newform := rec(
    name := ConcatenationString( form.name, "Int" ),
    operations := FormationOps,
    isFormation := true,
    isIntegrated := true );

    # new LocalResidual
    newform.LocalResidual := function( G, p )
        local R, L;
        R := FResidual( G, form );
        L := form.LocalResidual( G, p );
        if not IsList( L ) then
            return G.operations.MergedIgs( G, R, Igs( L ), false );
        else
            return [];
        fi;
    end;

    # add support
    if IsBound( form.support ) then
        newform.support := form.support;
    fi;

    return newform;
end;

#############################################################################
##
#F  FChangedSupport( form, support ) . . . . . formation with changed support
##
FChangedSupport := function( form, support )

    # check arguments
    if not IsFormation( form )  then
        Error("Sorry, need a formation to change support \n");
    fi;

    return form.operations.FChangedSupport( form, support );
end;

FormationOps.FChangedSupport := function( form, support )

    local newform;

    # set up new formation record
    newform := rec(
    name := form.name,
    operations := FormationOps,
    isFormation := true );

    # check integrated
    if IsBound( form.isIntegrated ) then
        newform.isIntegrated := form.isIntegrated;
    fi;

    # add support
    if IsBound( form.support ) then
        newform.support := Intersection( form.support, support );
    else
        newform.support := support;
    fi;

    # compute new LocalResidual 
    if IsBound( form.LocalResidual ) then
        newform.LocalResidual := function( G, p )
            if not p in newform.support then
                return [];
            else
                return form.LocalResidual( G, p );
            fi;
        end;
    fi;

    # now check whether there are some other functions bound
    if IsBound( form.FResidual ) then
        newform.FResidual := function( G )
            local R1, R2;
            R1 := form.FResidual( G );
            R2 := G.operations.PiResidual( G, newform.support );
            return G.operations.MergedIgs( G, R1, Igs( R2 ), false );
        end;
    fi;
   
    if IsBound( form.FNormalizer ) then
        newform.FNormalizer := function( G )
            local N;
            N := form.FNormalizer( G );
            return N.operations.HallSubgroup( N, newform.support );
        end;
    fi;

    if IsBound( form.FCoveringGroup ) then
        newform.FCoveringGroup := function( G )
            local N;
            N := form.FCoveringGroup( G );
            return N.operations.HallSubgroup( N, newform.support );
        end;
    fi;

    return newform;
end;

